from __future__ import annotations

import platform
from importlib.metadata import metadata
from tkinter import Toplevel
from tkinter.ttk import Frame, Label
from typing import TYPE_CHECKING

from .wrap_label import WrapLabel

if TYPE_CHECKING:
    from .app import TkApp


class TkAboutWindow(Toplevel):
    def __init__(self, app: TkApp) -> None:
        super().__init__(app)

        self.title("About")

        self.frame = Frame(self, padding=30)
        self.frame.pack(expand=True, fill="both")
        self.frame.grid_columnconfigure(0, weight=1)

        package = metadata("ollama-tk")

        self.name = Label(
            self.frame,
            font="Helvetica 12 bold",
            justify="center",
            text=f"OllamaTk v{package['Version']}",
        )
        self.name.grid(pady=(0, 15))

        self.description = WrapLabel(
            self.frame,
            justify="center",
            text=(
                f"By {package['Author']}\n\n"
                f"{package['Summary']}\n\n"
                f"Python version: {platform.python_version()}"
            ),
        )
        self.description.grid(pady=5)



import contextlib
import functools
import logging
import sys

from .app import TkApp
from .event_thread import EventThread
from .http import HTTPClient
from .logging import configure_logging


def suppress(*exceptions: type[BaseException]):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with contextlib.suppress(*exceptions):
                return func(*args, **kwargs)

        return wrapper

    return decorator


@suppress(KeyboardInterrupt)
def main() -> None:
    configure_logging()
    enable_windows_dpi_awareness()

    with EventThread() as event_thread:
        http = HTTPClient()
        http_task = event_thread.submit(http.run())

        app = TkApp(event_thread, http)
        app.listen_to_logs_from(logging.getLogger())

        try:
            app.mainloop()
        except BaseException:
            app.destroy()
            app.mainloop()
            raise


def enable_windows_dpi_awareness() -> None:
    if sys.platform == "win32":
        from ctypes import windll

        windll.shcore.SetProcessDpiAwareness(2)


if __name__ == "__main__":
    main()



import logging
from tkinter import Event, Menu, Tk
from tkinter.ttk import Frame

from .chat import TkChat, TkChatMenu
from .event_thread import EventThread
from .http import HTTPClient
from .logging import LogStore, TkAppLogHandler


class TkApp(Tk):
    def __init__(self, event_thread: EventThread, http: HTTPClient):
        super().__init__()

        self.event_thread = event_thread
        self.http = http
        self.logs = LogStore()

        self._connect_lifetime_with_event_thread(event_thread)

        self.title("Ollama Tk")
        self.geometry("560x670")
        self.option_add("*tearOff", False)

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.frame = TkChat(self)
        self.frame.grid(sticky="nesw")
        self.switch_menu(TkChatMenu(self))

        self.bind("<<Destroy>>", self._on_destroy)

    def switch_frame(self, frame: Frame) -> None:
        self.frame.destroy()
        self.frame = frame
        self.frame.grid(sticky="nesw")

    def switch_menu(self, menu: Menu | None) -> None:
        if menu is None:
            menu = Menu(self)

        self.configure(menu=menu)

    def listen_to_logs_from(self, logger: logging.Logger) -> TkAppLogHandler:
        handler = TkAppLogHandler(self)
        logger.addHandler(handler)
        return handler

    def _connect_lifetime_with_event_thread(self, event_thread: EventThread) -> None:
        # In our application we'll be running an asyncio event loop in
        # a separate thread. This event loop may try to run methods on
        # our GUI like event_generate(), which requires the GUI to be running.
        # If the GUI is destroyed first, it may cause a deadlock
        # in the other thread, preventing our program from exiting.
        # As such, we need to defer GUI destruction until the event thread
        # is finished.
        event_callback = lambda fut: self.event_generate("<<Destroy>>")
        event_thread.finished_fut.add_done_callback(event_callback)

    def destroy(self) -> None:
        self.event_thread.stop()

    def _on_destroy(self, event: Event) -> None:
        super().destroy()



from __future__ import annotations

import logging
from concurrent.futures import Future
from tkinter import Menu, Text
from tkinter.ttk import Button, Frame
from typing import TYPE_CHECKING, Any

from .about import TkAboutWindow
from .logging import TkLogWindow
from .messages import Message, TkMessageFrame, TkMessageList
from .settings import Settings, TkSettingsControls

if TYPE_CHECKING:
    from .app import TkApp

log = logging.getLogger(__name__)


class TkChat(Frame):
    chat_fut: Future | None

    def __init__(self, app: TkApp) -> None:
        super().__init__(app)

        self.app = app

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(1, weight=7)
        self.grid_rowconfigure(3, weight=1)

        self.settings = Settings()

        self.settings_controls = TkSettingsControls(self, self.settings)
        self.settings_controls.grid(row=0, column=0, sticky="e", padx=10, pady=(10, 0))

        self.message_list = TkMessageList(self)
        self.message_list.grid(row=1, column=0, sticky="nesw", padx=10, pady=(10, 0))

        self.live_controls = TkLiveControls(self)
        self.live_controls.grid(row=2, column=0, sticky="w", padx=10, pady=(10, 0))
        self.live_controls.grid_remove()  # Show later when necessary

        self.chat_controls = TkChatControls(self)
        self.chat_controls.grid(row=3, column=0, sticky="nesw", padx=10, pady=10)

        self.chat_fut = None

    def send_chat(self, *, source: TkMessageFrame | None) -> None:
        message = Message("assistant", "Waiting for response...")
        frame = self.message_list.add_message(message)

        coro = self.app.http.generate_chat_completion(
            target=frame,
            source=source,
            address=self.settings.ollama_address,
            model=self.settings.ollama_model,
            messages=self.message_list.dump(exclude=[frame]),
        )

        fut = self.chat_fut = self.app.event_thread.submit(coro)
        fut.add_done_callback(self._on_send_chat_done)

        self.settings_controls.disable()
        self.live_controls.show()
        self.chat_controls.disable()

    def _on_send_chat_done(self, fut: Future[Any]) -> None:
        self.settings_controls.enable()
        self.live_controls.grid_remove()
        self.chat_controls.enable()

        if not fut.cancelled() and fut.exception() is not None:
            log.exception("Error occurred while sending chat", exc_info=fut.exception())

    def maybe_get_models(self) -> None:
        # FIXME: update models any time address is changed
        if self.settings_controls.model["values"]:
            return

        coro = self.app.http.list_local_models(self.settings.ollama_address)
        fut = self.app.event_thread.submit(coro)
        fut.add_done_callback(self._on_maybe_get_models_done)

    def _on_maybe_get_models_done(self, fut: Future[list[str]]) -> None:
        if fut.cancelled():
            return
        elif fut.exception() is not None:
            return log.exception(
                "Error occurred while fetching available models",
                exc_info=fut.exception(),
            )

        self.settings_controls.model.configure(values=fut.result())


class TkLiveControls(Frame):
    def __init__(self, chat: TkChat) -> None:
        super().__init__(chat)

        self.chat = chat

        self.grid_columnconfigure(0, weight=1)

        self.cancel = Button(self, command=self.do_cancel, text="Cancel")
        self.cancel.grid(row=0, column=0)

    def do_cancel(self) -> None:
        if self.chat.chat_fut is not None:
            self.chat.chat_fut.cancel()
            self.cancel.state(["disabled"])

    def show(self) -> None:
        self.cancel.state(["!disabled"])
        self.grid()


class TkChatControls(Frame):
    def __init__(self, chat: TkChat) -> None:
        super().__init__(chat)

        self.chat = chat

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)

        self.text = Text(self, font="TkDefaultFont", width=50, height=4)
        self.text.grid(row=0, column=0, sticky="nesw", padx=(0, 20))
        self._init_text_bindings()

        self.buttons = TkChatButtons(self)
        self.buttons.grid(row=0, column=1, sticky="ns")

    def disable(self) -> None:
        self.text.configure(state="disabled")
        self.buttons.disable()

    def enable(self) -> None:
        self.text.configure(state="normal")
        self.buttons.enable()

    def _init_text_bindings(self) -> None:
        self.text.bind("<Shift-Return>", lambda event: self.text.insert("insert", ""))
        self.text.bind("<Return>", lambda event: self.buttons.do_send())


class TkChatButtons(Frame):
    def __init__(self, controls: TkChatControls) -> None:
        super().__init__(controls)

        self.controls = controls

        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure("0 1", weight=1)

        self.send_button = Button(self, command=self.do_send, text="Send")
        self.send_button.grid(row=0, column=0, sticky="new")

        self.clear_button = Button(self, command=self.do_clear, text="Clear")
        self.clear_button.grid(row=1, column=0, sticky="sew")

    def do_send(self) -> None:
        content = self.controls.text.get("1.0", "end").strip()
        if content == "":
            return

        message = Message("user", content)
        message = self.controls.chat.message_list.add_message(message)
        self.controls.text.delete("1.0", "end")
        self.controls.chat.send_chat(source=message)
        self.controls.chat.maybe_get_models()

    def do_clear(self) -> None:
        self.controls.chat.message_list.clear()

    def disable(self) -> None:
        self.send_button.state(["disabled"])
        self.clear_button.state(["disabled"])

    def enable(self) -> None:
        self.send_button.state(["!disabled"])
        self.clear_button.state(["!disabled"])


class TkChatMenu(Menu):
    def __init__(self, app: TkApp) -> None:
        super().__init__(app)
        self.app = app
        self.add_command(command=lambda: TkLogWindow(app), label="Logs")
        self.add_command(command=lambda: TkAboutWindow(app), label="About")



import asyncio
import json
from typing import Any, Literal, TypedDict, cast

import httpx

from .messages import Role, TkMessageFrame


# https://github.com/ollama/ollama/blob/main/docs/api.md#generate-a-chat-completion
class Message(TypedDict):
    role: Role
    content: str
    images: list[str] | None


class StreamingChat(TypedDict):
    model: str
    created_at: str
    message: Message
    done: Literal[False]


class DoneStreamingChat(TypedDict):
    model: str
    created_at: str
    done: Literal[True]
    total_duration: int  # in nanoseconds
    load_duration: int  # in nanoseconds
    prompt_eval_count: int
    prompt_eval_duration: int  # in nanoseconds
    eval_count: int
    eval_duration: int  # in nanoseconds


class HTTPClient:
    _client: httpx.AsyncClient | None

    def __init__(self) -> None:
        self._client = None

    @property
    def client(self) -> httpx.AsyncClient:
        if self._client is None:
            raise RuntimeError("HTTPClient is not running")
        return self._client

    async def run(self) -> None:
        if self._client is not None:
            raise RuntimeError("HTTPClient is already running")

        self._client = httpx.AsyncClient(timeout=10)
        try:
            async with self._client:
                await asyncio.get_running_loop().create_future()
        finally:
            self._client = None

    async def generate_chat_completion(
        self,
        *,
        target: TkMessageFrame,
        source: TkMessageFrame | None = None,
        address: httpx.URL | str,
        model: str,
        messages: list[dict[str, Any]],
    ) -> None:
        def show_error(message: str) -> None:
            if append_errors:
                target.message.content += f"...\n\n{message}"
            else:
                target.message.content = message
            target.refresh()

        def hide_messages() -> None:
            # Make sure a followup chat doesn't remember the failed messages
            target.message.hidden = True
            target.refresh()
            if source is not None:
                source.message.hidden = True
                source.refresh()

        address = httpx.URL(address).join("/api/chat")
        payload = {"model": model, "messages": messages}
        append_errors = False

        try:
            async with self.client.stream("POST", address, json=payload) as response:
                response.raise_for_status()

                append_errors = True
                target.message.content = ""
                target.refresh()

                async for line in response.aiter_lines():  # NOTE: what if this hangs?
                    data = cast(StreamingChat | DoneStreamingChat, json.loads(line))

                    error = data.get("error")
                    if error is not None:
                        raise RuntimeError(error)

                    if not data.get("done"):
                        data = cast(StreamingChat, data)
                        target.message.role = data["message"]["role"]
                        target.message.content += data["message"]["content"]
                    else:
                        data = cast(DoneStreamingChat, data)
                        # Nothing to do here really

                    target.refresh()

        except asyncio.CancelledError:
            show_error("(Response cancelled)")
            hide_messages()
            raise
        except httpx.ConnectError:
            show_error("Could not connect to the given address. Is the server running?")
            hide_messages()
        except httpx.HTTPStatusError as e:
            hide_messages()
            status = e.response.status_code
            phrase = e.response.reason_phrase

            if status == 400:
                show_error(f"{status} {phrase}. Did you select the model to run?")
            elif status == 404:
                show_error(
                    f"{status} {phrase}. Maybe your selected model does not exist?"
                )
            else:
                show_error(f"{status} {phrase}. Check logs for more details.")
                raise

        except Exception:
            # TODO: show more detailed error messages
            show_error("An unknown error occurred. Check logs for more details.")
            hide_messages()
            raise

    async def list_local_models(self, address: httpx.URL | str) -> list[str]:
        address = httpx.URL(address).join("/api/tags")
        response = await self.client.get(address)
        response.raise_for_status()
        return [model["name"] for model in response.json()["models"]]



from dataclasses import dataclass
from tkinter import Misc, StringVar
from tkinter.ttk import Combobox, Entry, Frame


@dataclass
class Settings:
    ollama_address: str = "http://localhost:11434"
    ollama_model: str = "llama3.1"


class TkSettingsControls(Frame):
    def __init__(self, parent: Misc, settings: Settings) -> None:
        super().__init__(parent)

        self.grid_columnconfigure("0 1", weight=1)

        self.settings = settings

        self.address_var = StringVar(self)
        self.address = Entry(self, textvariable=self.address_var)
        self.address.grid(row=0, column=0, padx=(0, 20))

        self.model_var = StringVar(self)
        self.model = Combobox(self, textvariable=self.model_var)
        self.model.grid(row=0, column=1)

        self.refresh()

        self.address_var.trace_add("write", self._on_address_var_write)
        self.model_var.trace_add("write", self._on_model_var_write)

    def refresh(self) -> None:
        self.address_var.set(self.settings.ollama_address)
        self.model_var.set(self.settings.ollama_model)

    def set_model_options(self, options: list[str]) -> None:
        self.model.configure(values=options)

    def disable(self) -> None:
        self.address.state(["disabled"])
        self.model.state(["disabled"])

    def enable(self) -> None:
        self.address.state(["!disabled"])
        self.model.state(["!disabled"])

    def _on_address_var_write(self, name1: str, name2: str, op: str) -> None:
        self.settings.ollama_address = self.address_var.get()

    def _on_model_var_write(self, name1: str, name2: str, op: str) -> None:
        self.settings.ollama_model = self.model_var.get()



# https://gist.github.com/thegamecracks/564bd55af973827f8a05b48f197d5c09
import asyncio
import concurrent.futures
import threading
from typing import Any, Coroutine, Self, TypeVar

T = TypeVar("T")


class EventThread(threading.Thread):
    """Runs an asyncio event loop in a separate thread.
    Starting and stopping can be done with :meth:`start()` and :meth:`stop()`,
    but for reliability, you should use the class in a context manager instead::
        >>> with EventThread() as event_thread:
        ...     loop = event_thread.loop
        ...     future = asyncio.run_coroutine_threadsafe(my_task(), loop)
        ...     result = future.result()
        >>> # Event loop will be cleaned up before exiting
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.loop_fut = concurrent.futures.Future()
        self.stop_fut = concurrent.futures.Future()
        self.finished_fut = concurrent.futures.Future()

    def __enter__(self) -> Self:
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, tb) -> None:
        self.stop()
        self.join()

    @property
    def loop(self) -> asyncio.AbstractEventLoop:
        return self.loop_fut.result()

    def run(self) -> None:
        try:
            asyncio.run(self._run_forever())
        finally:
            self.finished_fut.set_result(None)

    def submit(self, coro: Coroutine[Any, Any, T]) -> concurrent.futures.Future[T]:
        return asyncio.run_coroutine_threadsafe(coro, self.loop)

    def stop(self) -> None:
        try:
            self.stop_fut.set_result(None)
        except concurrent.futures.InvalidStateError:
            pass

    async def _run_forever(self) -> None:
        self.loop_fut.set_result(asyncio.get_running_loop())
        await asyncio.wrap_future(self.stop_fut)



# https://gist.github.com/thegamecracks/5595dad631ec50bdc021f945054e86fb
from tkinter import Event
from tkinter.ttk import Label


class WrapLabel(Label):
    """A label that automatically wraps its text when resized.

    When using this label, the geometry manager must be configured
    to resize the widget in some way, otherwise the width will
    default to the text's max length.
    This can mean using ``.pack(fill="both", expand=True)``, or
    ``.place(relwidth=1, relheight=1)``, or by configuring grid
    weights appropriately.

    ``minwidth=`` can be specified to prevent wrapping under
    a certain width which can significantly improve performance
    for longer lines of text.

    As a limitation of the current implementation, this class can only
    match its bounding box when used with the grid geometry manager.
    When using other geometry managers like pack or place, WrapLabel will
    always wrap to the window's entire width, regardless of the actual
    bounding box allocated for them. As a result, the text may appear clipped.

    If you use those geometry managers, it is recommended to grid / pack
    each WrapLabel inside a frame so it can calculate the wrap length from
    that frame's width. For example::

        # Instead of packing the label directly:
        label = WrapLabel(root, text="Some long piece of text")
        label.pack(fill="both", expand=True, padx=200, pady=200)

        # Pack a container frame and have the label expand inside it:
        frame = Frame(root)
        frame.pack(fill="both", expand=True, padx=200, pady=200)
        label = WrapLabel(frame, text="Some long piece of text")
        label.pack(fill="both", expand=True)

    In effect, the container frame serves as a way to compute the true
    bounding box allocated by the geometry manager, helping the label
    wrap to the correct width.

    """

    def __init__(self, *args, minwidth: int = 150, **kwargs):
        super().__init__(*args, **kwargs)
        self.minwidth = minwidth
        self.bind("<Configure>", self.__on_configure)

    def __on_configure(self, event: Event):
        width = max(self.minwidth, self.__get_width())
        if width != 1:  # Prevent wrapping on initial configuration
            self.configure(wraplength=width)

    def __get_width(self) -> int:
        if self.winfo_manager() == "grid":
            # Wrap to the bounding box reserved for the label
            # instead of the container's full width.
            # Not doing this might lead to clipped text.
            options = self.grid_info()
            bbox = self.master.grid_bbox(options["column"], options["row"])
            if bbox is None:
                return 1

            width = bbox[2]
            if isinstance(options["padx"], int):
                padx = options["padx"] * 2
            else:
                padx = sum(options["padx"])

            return width - padx
        else:
            return self.master.winfo_width()



from __future__ import annotations

import logging
from contextlib import contextmanager, nullcontext
from tkinter import Text, Toplevel
from tkinter.ttk import Button, Scrollbar
from typing import TYPE_CHECKING, Any, Callable, Iterator, Literal, Self

if TYPE_CHECKING:
    from .app import TkApp

LogEventType = Literal["clear", "insert"]


def configure_logging() -> None:
    logging.basicConfig(level=logging.INFO)


class TkAppLogHandler(logging.Handler):
    def __init__(self, app: TkApp) -> None:
        super().__init__()
        self.app = app

    def emit(self, record: logging.LogRecord) -> None:
        message = self.format(record)
        self.app.logs.append(message)


class LogStore:
    callbacks: list[Callable[[LogEventType, str], Any]]
    _messages: list[str]

    def __init__(self) -> None:
        self.callbacks = []
        self._messages = []

    def __iter__(self) -> Iterator[str]:
        return iter(self._messages.copy())

    def append(self, message: str) -> None:
        self._messages.append(message)
        self._notify("insert", message)

    def clear(self) -> None:
        if len(self._messages) < 1:
            return

        self._messages.clear()
        self._notify("clear", "")

    def _notify(self, type: LogEventType, message: str) -> None:
        for callback in self.callbacks.copy():
            callback(type, message)


class TkLogWindow(Toplevel):
    def __init__(self, app: TkApp) -> None:
        super().__init__(app)

        self.app = app

        self.title("Logs")
        self.geometry("800x550")

        self.clear = Button(self, command=self.do_clear, text="Clear")
        self.clear.pack(side="bottom", anchor="e", padx=(0, 10), pady=10)

        self.text = Text(
            self,
            font="TkDefaultFont",
            width=0,
            height=0,
            state="disabled",
        )
        self.text.pack(
            side="left",
            expand=True,
            fill="both",
            padx=(10, 0),
            pady=(10, 0),
        )

        self.scrollbar = Scrollbar(self, command=self.text.yview)
        self.text.configure(yscrollcommand=self.scrollbar.set)
        self.scrollbar.pack(expand=True, fill="y", padx=(0, 10), pady=(10, 0))

        self.refresh()
        self.app.logs.callbacks.append(self._on_log_update)

    def refresh(self) -> None:
        with self.unlock_text():
            self.text.delete("1.0", "end")
            for message in self.app.logs:
                self.text.insert("end", message + "\n")

    def do_clear(self) -> None:
        self.app.logs.clear()

    def destroy(self) -> None:
        self.app.logs.callbacks.remove(self._on_log_update)
        super().destroy()

    def _on_log_update(self, type: LogEventType, message: str) -> None:
        with self.unlock_text():
            if type == "clear":
                self.text.delete("1.0", "end")
            elif type == "insert":
                self.text.insert("end", message + "\n")

    @contextmanager
    def unlock_text(self, *, autoscroll: bool = True) -> Iterator[Self]:
        scroll_manager = self.autoscroll() if autoscroll else nullcontext()
        with scroll_manager:
            self.text.configure(state="normal")
            try:
                yield self
            finally:
                self.text.configure(state="disabled")

    @contextmanager
    def autoscroll(self) -> Iterator[Self]:
        scrolled_to_bottom = self.text.yview()[1] == 1
        try:
            yield self
        finally:
            if scrolled_to_bottom:
                self.text.yview_moveto(1)



from __future__ import annotations

import importlib.resources
from dataclasses import dataclass
from tkinter import Event, PhotoImage
from tkinter.ttk import Frame, Label
from typing import TYPE_CHECKING, Any, Collection, Literal

from .scrollable_frame import ScrollableFrame
from .wrap_label import WrapLabel

if TYPE_CHECKING:
    from .chat import TkChat

Role = Literal["system", "user", "assistant", "tool"]


@dataclass
class Message:
    role: Role
    content: str
    hidden: bool = False

    def dump(self) -> dict[str, Any]:
        return {"role": self.role, "content": self.content}


class TkMessageFrame(Frame):
    def __init__(
        self,
        message_list: TkMessageList,
        message: Message,
        *,
        side: Literal["left", "right"],
    ) -> None:
        super().__init__(message_list.inner)

        self.message = message
        self.message_list = message_list

        self.grid_columnconfigure(1, weight=1)
        self.grid_rowconfigure(1, weight=1)

        left = side == "left"
        anchor = "w" if left else "e"

        self.role_label = Label(self, anchor=anchor, justify=side)
        self.role_label.grid(row=0, column=1, sticky="ew")

        self.role_icon = Label(self)
        self.role_icon.grid(
            row=1,
            column=0 if left else 2,
            sticky="n",
            padx=(0, 5) if left else (5, 0),
        )

        self.content_label = WrapLabel(self, anchor=anchor, justify=side)
        self.content_label.bind("<1>", self._on_content_label_click)
        self.content_label.grid(row=1, column=1, sticky="nesw")

        self.refresh()

    def refresh(self) -> None:
        role = self.message.role.title() + " (hidden)" * self.message.hidden
        self.role_label.configure(text=role)
        self.content_label.configure(text=self.message.content)

        if self.message.role == "user":
            self.role_icon.configure(image=self.message_list.icons["user"])
        else:
            self.role_icon.configure(image=self.message_list.icons["assistant"])

    def destroy(self) -> None:
        super().destroy()
        self.message_list.messages.remove(self)

    def _on_content_label_click(self, event: Event) -> None:
        self.clipboard_clear()
        self.clipboard_append(self.content_label["text"])


class TkMessageList(ScrollableFrame):
    messages: list[TkMessageFrame]

    def __init__(self, chat: TkChat) -> None:
        super().__init__(chat, autoscroll=True, yscroll=True)

        self.chat = chat
        self.messages = []

        self.inner.grid_columnconfigure(0, weight=1)

        self.icons = load_message_icons()

    def add_message(self, message: Message) -> TkMessageFrame:
        side = "right" if message.role == "user" else "left"
        frame = TkMessageFrame(self, message, side=side)
        frame.grid(row=len(self.messages), column=0, sticky="ew")
        self.messages.append(frame)
        return frame

    def refresh(self) -> None:
        for message in self.messages:
            message.refresh()

    def clear(self) -> None:
        for message in self.messages.copy():
            message.destroy()  # NOTE: this is O(n^2)

    def dump(
        self,
        *,
        exclude: Collection[TkMessageFrame] = (),
        include_hidden: bool = False,
    ) -> list[dict[str, Any]]:
        return [
            frame.message.dump()
            for frame in self.messages
            if (include_hidden or not frame.message.hidden) and frame not in exclude
        ]


def load_message_icons() -> dict[str, PhotoImage]:
    icons = importlib.resources.files("ollamatk.icons")
    return {
        "assistant": PhotoImage(data=icons.joinpath("assistant.png").read_bytes()),
        "user": PhotoImage(data=icons.joinpath("user.png").read_bytes()),
    }



from .about import TkAboutWindow
from .app import TkApp
from .chat import TkChat, TkChatButtons, TkChatControls, TkLiveControls, TkChatMenu
from .event_thread import EventThread
from .http import HTTPClient
from .logging import LogStore, TkAppLogHandler, TkLogWindow, configure_logging
from .messages import Message, TkMessageFrame, TkMessageList, load_message_icons
from .scrollable_frame import ScrollableFrame
from .settings import Settings, TkSettingsControls
from .wrap_label import WrapLabel
